import os

WAIT_FOR_REQUEST_IN_EMPTY_QUEUE_DELAY = 1.5
MAX_REQUEST_QUEUE_LEN = 30000
REQUEST_MAX_FREQUENCY = 0.1

METRICS_SERVER_HOST = os.environ.get('METRICS_SERVER_HOST', '127.0.0.1')
METRICS_SERVER_PORT = int(os.environ.get('METRICS_SERVER_PORT', 8008))
